/****************************************************************************\


  Block deleteion Algorithm - should work with ALL posible cases.


  Written by Harvey J "Max" Thompson (max@cs.warwick.ac.uk)

  Maths added by M. A. "Slimo" Fincher (mauls@cu.warwick.ac.uk)

Definitions
-----------

Atom      -	One square, or character occupying one space.

Colour	  -	The particular character used for an atom.

Hole	  -	Absence of any atom in a particular space, colour
		is black.

Connected -	A is connected to B if A is vertically or horizontally
		adjacent to B.

Piece	  -	A collection of like coloured pieces, which are connected.

Set	  -	A group of connected pieces of different colours.

Block	  -	A set which has a continous border, which is rectangular
		in shape. No piece may have a protruding atom over this
		boundary, however the block can have holes inside this
		boundary.



Axioms
------

1) To complete a block, the final atom added will always be on the border.

2) A block is of minimum height, or width of two.

\****************************************************************************/

#ifndef DEBUG
#include <curses.h>
#else
#include <stdio.h>
extern char block[5][10];
#endif

#include "check.h"
				/* Wether an entry has been computed already */
static int done[MAX_DOWN][MAX_LEFT]; /* Bits 0-31 for right */

static QUEUE *head = 0;		/* No head */
static QUEUE *tail = 0;		/* No tail */

/****************************************************************************\
  Get the colour of a x,y coordinate in the block-hole orientation that
  I'm gonna work with. That is x along the horizontal, y along the vertical.
  Origin x=0,y=0 is at the top-left. (The screen displays this at
  90 degrees clockwise).
\****************************************************************************/

static char get_colour ( x, y )
    int x;
    int y;
{
    char ch;

    if ( x < MIN_X || x > MAX_X || y < MIN_Y || y > MAX_Y )
	return ( BLACK );
    else {
#ifndef DEBUG
	move ( x, (79-y) );
	ch = inch ();
	if ( ch == 'o' )
	    return ( BLACK );
	else
	    return ( ch );
#else

	return ( block[y][x] );
#endif
    }
}

/****************************************************************************\
  Check a block starting at x,y of depth d, and left-width l, right-width
  r has a complete and correctly formed boundary. Also return whether it
  is possible to iterate in the left, right and/or back directions.

  Returns;

  Bit   Meaning if UNSET
  0     an intact boundary been found here

  1     continuation on the left is possible
  2     continuation on the right is possible
  3     continuation on the back is possible (at this left or right)

  ie. A single atom has depth = 0, left = 0, right = 0

\****************************************************************************/

static int check_block ( x, y, l, r, d )
    int x;			/* X-coord to start at */
    int y;			/* Y-coord to start at */
    int l;			/* left width */
    int r;			/* right width */
    int d;			/* depth */
{
    int  bx,			/* Boundary x-coord */
         by;			/* Boundary y-coord */
    char bc,			/* Border colour */
         oc;			/* Outside colour */
    int  flag = 0;		/* Return flag - asumme expandable */
				/* in all directions and border found */

				/* Check nothing nasty infront left */
    bc = get_colour ( x-l, y );
    oc = get_colour ( x-l, y+1 );

    if ( oc != BLACK && oc == bc )
	flag |= 1+2;		/* No border here and can't continue left */

				/* Check nothing nasty infront right */
    bc = get_colour ( x+r, y );
    oc = get_colour ( x+r, y+1 );

    if ( oc != BLACK && oc == bc )
	flag |= 1+4;		/* No border here and can't continue right */

				/* Check left and right boundary */
    for ( by = y-d; by <= y; by++ ) {
				/* Check left bound */
	bc = get_colour ( x-l, by );
	oc = get_colour ( x-l-1, by );

	if ( bc == BLACK ) {
	    flag |= 1;		/* No border here! */
	    if ( by == y )
		flag |= 2;	/* Can't continue left */
	    else if ( by == y-d )
		flag |= 8;	/* Can't continue back */
	}
				/* Can't have a border with protrusions */
	if ( oc != BLACK && bc == oc )
	    flag |= 1;
				/* If all flags set we might as well frog */
				/* off */
	if ( flag == 15 )
	    return ( 15 );
				/* Check right bound */
	bc = get_colour ( x+r, by );
	oc = get_colour ( x+r+1, by );

	if ( bc == BLACK ) {
	    flag |= 1;		/* No border here! */
	    if ( by == y )
		flag |= 4;	/* Can't continue right */
	    else if ( by == y-d )
		flag |= 8;	/* Can't continue back */
	}
				/* Can't have a border with protrusions */
	if ( oc != BLACK && bc == oc )
	    flag |= 1;
				/* If all flags set we might as well frog */
				/* off */
	if ( flag == 15 )
	    return ( 15 );
    }

				/* Check back bound */
    for ( bx = x-l; bx <= x+r; bx++ ) {
	bc = get_colour ( bx, y-d );
	oc = get_colour ( bx, y-d-1 );
				/* No spaces in the border please */
	if ( bc == BLACK )
	    flag |= 1;		/* No border here! */
				/* No protrosions */
	if ( oc != BLACK && bc == oc )
	    flag |= 1;
				/* If all flags set we might as well frog */
				/* off */
	if ( flag == 15 )
	    return ( 15 );
				/* We assume the front is ok */
    }
    return ( flag );
}

/****************************************************************************\
  Intialise queue to be empty
\****************************************************************************/
static void init_queue ()
{
    head = 0;
    tail = 0;
    bzero ( done, sizeof ( done ) );
}

/****************************************************************************\
  Add an entry to the queue, making sure that the entry has never been
  added before.

  Without this pruining technique the O(ml,mr,md)  = (ml+md)!  + (mr+md)!
                                                      ------       ------
						     ml! . md!   mr!. md!

  However with it - the process is reduced to;

  O(ml,mr,md) = ml * mr * md

  Which is a VAST saving...

  A test run for example - without pruning O(6,3,4) == 4399
                           with    pruning O(6,3,4) == 72

\****************************************************************************/
static void add_queue ( l, r, d, flag )
    int l,
	r,
	d,
	flag;
{
    QUEUE *new;

#ifndef NOPRUNE
				/* Have we done this entry already? */
    if ( done[d][l] & (1 << r))
	return;			/* Yeah so fish off */
    done[d][l] |= (1 << r);	/* Naah we aint - mark as if done */
#endif

    if ( ( new = (QUEUE *) malloc ( sizeof ( QUEUE ) ) ) == 0 ) {
	fprintf ( stderr, "Couldn't allocate memory for queue.\n" );
	exit ( 1 );
    }
				/* Initialise entry */
    new->left = l;
    new->right = r;
    new->depth = d;
    new->flag = flag;
				/* Add to the end of the list */
    if ( ! tail ) {		/* If no tail - new queue */
	tail = new;
	head = new;
	new->next = 0;
    } else {
	tail->next = new;	/* The old tail has us after */
	new->next = 0;		/* We havn't however! */
	tail = new;		/* Since we're the tail */
    }
}

/****************************************************************************\
  Get an entry from the head of the queue, and delete it.
  Return 1 if queue non-empty
\****************************************************************************/
static int get_queue ( l, r, d, flag )
    int *l,
	*r,
	*d,
	*flag;
{
    QUEUE *this = head;		/* Remeber the head */

    if ( ! head )
	return 0;
				/* Set return thingies */
    *l = head->left;
    *r = head->right;
    *d = head->depth;
    *flag = head->flag;
				/* Reset head to point to next thingy */
    if ( ( head = head->next) == 0 )
	tail = 0;
    free ( this );		/* Free this thing man! */
    return 1;
}

/****************************************************************************\
  Find the largest block starting at x, y return left, right and depth
  to the integers pointed to in the arguments.

  Return 1 if a block needs to be deleted,
         0 if none.
\****************************************************************************/

int find_block ( x, y, ml, mr, md )
    int x,
	y,
	*ml,
	*mr,
	*md;
{
    int    flag;		/* Current flag */
    int    cl = 0,		/* Current left */
           cr = 0,		/* Current right */
           cd = 1;		/* Current depth */
    int    found = 0;		/* Found a block to delete */
#ifdef DEBUG
    int    i = 1;
#endif
				/* No entries have been done */
    init_queue ();

    *ml = 0;			/* Initialise as empty */
    *mr = 0;
    *md = 0;

    for (;;) {
				/* Check the block we fired on, and */
				/* the one below */
	flag = check_block ( x, y, cl, cr, cd );

#ifdef DEBUG
	printf ( "%d: cl = %d, cr = %d, cd = %d, flag = %d%d%d%d\n",
		i++,
		cl, cr, cd,
		(flag & 1),
		((flag & 2) >> 1),
		((flag & 4) >> 2),
		((flag & 8) >> 3) );
#endif
				/* A block has been found, set */
				/* flag true, and set maximum */
				/* limits (another may be found) */
	if ( ! (flag & 1) ) {
	    *ml = cl;
	    *mr = cr;
	    *md = cd;
	    found = 1;
	}
				/* We can continue left ... add */
				/* to the left component and */
				/* process later. */
	if ( ! (flag & 2) )
	    add_queue ( cl+1, cr, cd, flag );
				/* We can continue right */
	if ( ! (flag & 4) )
	    add_queue ( cl, cr+1, cd, flag );
				/* We can continue back */
	if ( ! (flag & 8) )
	    add_queue ( cl, cr, cd+1, flag );

				/* Get the next processing item from */
				/* queue - if non - then we have */
				/* nothing more to do */
	if ( ! get_queue ( &cl, &cr, &cd, &flag ) )
	    break;
    }
				/* Return if we have found a block */
    return ( found );
}




